//---------------------------------------------------------------------------
//
// Evj        Model the evolution of DNA binding site recognition
//
// Module:    Simulator, the class that performs the simulation
// Author:    Paul C. Anagnostopoulos
// Created:   2 April 2005
//
// This computer program fits the description in the U.S. Copyright Act
// of a "United States Government work." It was written as a part of
// my/our official duties as a Government employee. This means it cannot
// be copyrighted. The program is freely available for use without a
// copyright notice, and there are no restrictions on its use, now or
// subsequently. We retain no rights in the program. Please acknowledge
// the source of this code in any releases you make.
//
//---------------------------------------------------------------------------


import java.lang.management.*;
import java.util.Arrays;
import java.util.Random;

//                      Instance Variables
//                      -------- ---------


public final class Simulator implements Runnable
{
  // Here we have the simulation parameter object, along with some
  // additional values derived from the parameters.
  
  private Param param;
  private int thresholdWidth;
  private static final int genePadding = 0;
  private int genomeSize;
  private double mutations;       // Mutations per genome per generation.
  private Random generator;       // Random number generator.

  // These variables maintain the information that represents the
  // simulation.
  
  private int genePos = 0;        // Position of recognizer gene,
  private int thresholdPos;       // and weight threshold.
  private int availablePos,       // Position of binding site portion of
              availableLen;       // the chromosome, and its length.
  private int sitePos[];          // Array of binding site positions.
  private boolean siteInd[];      // Array of binding site indicators.
  private Creature bug[];         // Array of creatures.

  // These variables maintain the state of the simulation.

  private Thread thread;                  // Thread that runs the simulation.
  private int generation = 0;             // Current generation.
  private int speed = Evj.pause;          // Speed at which to simulate.
  private int updateFreq;                 // Display update frequency.
  private int pausedPerfect = 0;          // Have paused on perfect creature.
  private int pausedConverged = 0;        // Have paused on convergence
                                          //   (Rsequence >= Rfrequency). 
  private int pausedPerfecged = 0;        // Have paused on both.
  private int deaths = 0;                 // Total deaths.
  private int selectionDeaths = 0;        // Deaths by selection.

//                      Constructor and Initialization
//                      ----------- --- --------------


  // The constructor accepts a parameter object and initializes the
  // simulation.
  
  public Simulator (Param param, int freq)
  {
    this.param = param;
    updateFreq = freq;
    setupNewSimulation();
    thread = new Thread(this);
    thread.setDaemon(true);
    thread.start();
///ThreadMXBean tbe  =  ManagementFactory.getThreadMXBean();
    }

  // This method sets up a new simulation.

  private void setupNewSimulation ()
  {
    // Initialize a few things that are directly related to the
    // parameters.
    
    thresholdWidth = param.getInt(Param.weightWidth);
    genomeSize = param.getInt(Param.potentialSites) +
                 param.getInt(Param.siteWidth) - 1;
    generator = new Random(param.getInt(Param.seed));

    // Calculate some statistics derived from the parameters.

    Stats.calculateDerived(param);
    
    // Calculate the position of the threshold.

    thresholdPos = genePos +
                   param.getInt(Param.siteWidth) *
                     4 * param.getInt(Param.weightWidth);

    // Calculate the position and length of the binding site portion
    // of the chromosome.

    availablePos = thresholdPos + thresholdWidth + genePadding;
    availableLen = genomeSize - availablePos;

    // Determine the binding sites.

    determineBindingSites();

    // Determine the mutations per genome per generation.

    determineMutations();
    
    // Create and initialize the creatures.

    populate();
  }

  // Populate the landscape with the requested number of creatures.
  // For each founder, we create a new creature and then the
  // appropriate number of clones. The last founder is used to clone
  // any extra creatures.
  
  private void populate ()
  {
    int pop = param.getInt(Param.population),
        founders = (param.getBoolean(Param.useFounder)) ?
                     param.getInt(Param.founders) : pop,
        clones = pop / founders - 1;
    
    bug = new Creature[pop];
    Creature.idCounter = 0;

    for (int b = -1, f = 1; f <= founders; ++f) {
      bug[++b] = new Creature(param, genomeSize, genePos, thresholdPos,
                              thresholdWidth, availablePos, generator);
      int founder = b;
      if (f == founders) clones = pop - b - 1;
      for (int c = 1; c <= clones; ++c)
        bug[++b] = new Creature(bug[founder]);
    }
  }
  
  // This method determines the binding site positions. Binding sites
  // appear in the available portion of the chromosome, following the
  // recognizer gene.

  private void determineBindingSites ()
  {
    // Create the site position array and fill it using an evenly
    // spaced array of binding sites or randomly chosen ones.

    sitePos = new int[param.getInt(Param.sites)];
    switch (param.getCode(Param.placement)) {
      case Param.PLACE_ARRAYED:
        arrayedBindingSites();
        break;

      case Param.PLACE_RANDOM: case Param.PLACE_RANDOM_OVERLAP:
        if (!randomBindingSites()) {
          Msg.warning(null, "Could not randomly allocate " +
                      param.getInt(Param.sites) +
                      " " + param.getInt(Param.siteWidth) +
                      "-base binding sites.\nAllocating a " +
                      "regularly-spaced array of binding sites instead.");
          arrayedBindingSites();
        }
        break;
    }

    // Sort the site positions in order. This makes various display
    // algorithms easier.

    Arrays.sort(sitePos);
    
    // Now create the site indicator vector and set only those
    // bits that correspond to the beginning of a site.

    siteInd = new boolean[genomeSize];
    Arrays.fill(siteInd, false);
    for (int i = 0; i < sitePos.length; ++i) siteInd[sitePos[i]] = true;
  }

  // This method allocates the binding sites in an evenly-spaced array
  // throughout the available chromosome space. The stride specifies
  // the number of bases between sites.

  private void arrayedBindingSites ()
  {
    int stride = availableLen / param.getInt(Param.sites);
    
    for (int i = 0; i < sitePos.length; ++i)
      sitePos[i] = availablePos + i * stride;
  }

  // This method allocates the binding sites randomly.

  private boolean randomBindingSites ()
  {
    int width = param.getInt(Param.siteWidth);
    boolean disjoint = param.getCode(Param.placement) == Param.PLACE_RANDOM;

    // Determine the bases available for sites.

    int available = availableLen - width + 1;

    // Set up a boolean array to indicate where sites are allocated.

    boolean allocated[] = new boolean[genomeSize];
    Arrays.fill(allocated, false);

    // Make 3*available attempts to allocate sites. Pick a random
    // position, then check to see if it already has a site. If not,
    // allocate a site there and mark it in the indicator array.

    int count = 0;
    for (int n = 1; count < sitePos.length && n <= 3 * available; ++n) {
      int p = availablePos + generator.nextInt(available);
      if (disjoint) {
        if (allocated[p] || allocated[p+width-1]) continue;
        sitePos[count] = p;
        Arrays.fill(allocated, p, p + width, true);
      } else {
        if (allocated[p]) continue;
        sitePos[count] = p;
        allocated[p] = true;
      }
      ++count;
    }

    // Return boolean, true if all sites were allocated.

    return count == sitePos.length;
  }

  private void determineMutations ()
  {
    switch (param.getCode(Param.mutationMethod)) {
      case Param.MUTATION_PER_GENOME:
        mutations = param.getInt(Param.mutationsPerGenome);
        break;
        
      case Param.MUTATION_PER_BASES:
        mutations = (double)genomeSize /
                    param.getInt(Param.mutationsPerBases);
        break;
    }
  }

//                      Simulation Loop
//                      ---------- ----


  public void run ()
  {
    update();

    for (;;) {
      if (speed == Evj.pause2) ;
      
      else if (generation >= param.getInt(Param.cycles)) {
        speed = Evj.pause;
        update();

      } else if (speed != Evj.pause) {
        ++generation;
        if (generation > 1) {
          reproduce();
          mutate();
        }
        evaluate();
        if (!param.getBoolean(Param.selection)) randomizeSort();
        order();
        checkPauses();
        if (speed == Evj.pause || generation % updateFreq == 0) update();
      }

      try {
        switch (speed) {
        case Evj.pause:   speed = Evj.pause2; // Fall through.
        case Evj.pause2:  thread.sleep(250); break;
        case Evj.quit:    break;
        default:          thread.sleep(1000 / (1 << speed-1)); break;
        }
      } catch (InterruptedException e) { }

      if (speed == Evj.quit) break;
    }
  }

  // This method checks for all the various situations that should
  // cause the simulation to pause. The paused* variables are set to 1
  // to indicate pausing on this cycle. The speed is set to "paused."
  
  private void checkPauses ()
  {
    double bestRseq = Stats.Rsequence(param, sitePos, bug[0].chromosome);
    
    if (pausedPerfecged == 0 && param.getBoolean(Param.perfecgence) &&
        bug[0].mistakes == 0 && bestRseq >= param.Rfreq) {
      pausedPerfecged = pausedPerfect = pausedConverged = 1;
      speed = Evj.pause;
      
    } else if (pausedPerfect == 0 && param.getBoolean(Param.perfection) &&
               bug[0].mistakes == 0) {
      pausedPerfect = 1;
      speed = Evj.pause;
    
    } else if (pausedConverged == 0 && param.getBoolean(Param.convergence) &&
               bestRseq >= param.Rfreq) {
      pausedConverged = 1;
      speed = Evj.pause;
    
    } else if (speed == Evj.step)
      speed = Evj.pause;
  }
  
  private void update ()
  {
    double bestRseq = Stats.Rsequence(param, sitePos, bug[0].chromosome);
    
    // Fill the nexus with all its information.
    
    Nexus nexus = new Nexus();
    nexus.title          = param.getString(Param.title);
    nexus.generation     = generation;
    nexus.cycles         = param.getInt(Param.cycles);
    nexus.pausePerfect   = pausedPerfect == 1;
    nexus.pauseConverged = pausedConverged == 1;
    nexus.population     = bug.length;
    nexus.potentialSites = param.getInt(Param.potentialSites);
    nexus.sites          = param.getInt(Param.sites);
    nexus.siteWidth      = param.getInt(Param.siteWidth);
    nexus.mutations      = mutations;
    nexus.selection      = param.getBoolean(Param.selection);
    nexus.Rfreq          = param.Rfreq;
    nexus.bestID         = bug[0].id;
    nexus.worstID        = bug[bug.length-1].id;
    nexus.bestAge        = bug[0].age;
    nexus.worstAge       = bug[bug.length-1].age;
    nexus.bestMistakes   = (generation == 0) ? -1
                                             : bug[0].mistakes;
    nexus.worstMistakes  = (generation == 0) ? -1
                                             : bug[bug.length-1].mistakes;
    nexus.bestRseq       = bestRseq;
    nexus.worstRseq      = Stats.Rsequence(param, sitePos,
                                           bug[bug.length-1].chromosome);
    nexus.perfection     = param.getBoolean(Param.perfection);
    nexus.convergence    = param.getBoolean(Param.convergence);
    nexus.perfecgence    = param.getBoolean(Param.perfecgence);

    nexus.weightWidth    = param.getInt(Param.weightWidth);
    nexus.useFounder     = param.getBoolean(Param.useFounder);
    nexus.founders       = param.getInt(Param.founders);
    nexus.genePos        = genePos;
    nexus.thresholdPos   = thresholdPos;
    nexus.availablePos   = availablePos;
    nexus.sitePos        = sitePos;
    
    nexus.births         = Creature.idCounter;
    nexus.deaths         = deaths;
    nexus.selectionDeaths= selectionDeaths;
    nexus.E_Hnb          = param.E_Hnb;
    
    nexus.chromosome     = (byte[])(bug[0].chromosome.clone());
    nexus.threshold      = bug[0].threshold;
    nexus.minValuation   = bug[0].getMinValuation();
    nexus.maxValuation   = bug[0].getMaxValuation();
    nexus.posInfo        = bug[0].getPosInfo(generation, siteInd);
    
    // Remember that we've paused on the various conditions by setting
    // the paused* variables to 2.
    
    if (nexus.pausePerfect)   pausedPerfect = 2;
    if (nexus.pauseConverged) pausedConverged = 2;
    if (pausedPerfecged == 1) pausedPerfecged = 2;

    // Send the nexus to the GUI.
    
    Evj.updateGUI(nexus);
  }

  // Accept the simulation speed from the GUI.
  
  public void setSpeed (int speed)
  {
    this.speed = speed;
  }

  // Accept the display update frequency from the GUI.

  public void setUpdateFreq (int freq)
  {
    updateFreq = freq;
  }

//                      Reproduction Phase
//                      ------------ -----


  // This method performs the reproduction step. The best half of the
  // bugs are the parents, and the worst half are killed off and become
  // the children. If there are any tie scores in the middle of the
  // sorted array, they are handled according to the tie parameter.

  private void reproduce ()
  {
    char tie = param.getCode(Param.tie);
    int worstBest = bug[bug.length/2-1].mistakes;

    for (int p = 0; p <= bug.length/2 - 1; ++p) {
      int c = bug.length - p - 1;

      if (bug[p].mistakes == bug[c].mistakes) {
        switch (tie) {                      // Two creatures have tie scores:
          case Param.TIE_BOTH:                //   Both survive.
            break;
          case Param.TIE_RANDOM:              //   A random one survives.
            if (generator.nextBoolean())
              killAndClone(c, p, worstBest);
            else
              killAndClone(p, c, worstBest);
            break;
          case Param.TIE_FIRST:               //    The parent surives.
            killAndClone(c, p, worstBest);
            break;
        }

      } else {
        killAndClone(c, p, worstBest);
      }
    }
  }

  private void killAndClone (int c, int p, int worstBest)
  {
    ++deaths;
    if (bug[c].mistakes > worstBest) ++selectionDeaths;
    bug[c].killAndClone(bug[p]);
  }

//                      Mutation Phase
//                      -------- -----


  // This method mutates the chromosome of each bug the number of times
  // calculated during initialization. If there is a fractional
  // mutation, we perform it with the appropriate probability.

  private void mutate ()
  {
    int mu = (int)mutations;
    double frac = mutations - (double)mu;

    switch (param.getCode(Param.mutationMethod)) {
      case Param.MUTATION_PER_GENOME:
        for (int b = 0; b < bug.length; ++b) {
          for (int hit = 1; hit <= mutations; ++hit) bug[b].mutate();
        }
        break;

      case Param.MUTATION_PER_BASES:
        for (int b = 0; b < bug.length; ++b) {
          for (int hit = 1; hit <= mu; ++hit) bug[b].mutate();
          if (generator.nextDouble() < frac) bug[b].mutate();
        }
        break;
    }
  }

//                      Evaluation Phase
//                      ---------- -----


  // This method "evaluates" each bug by computing a score based on
  // how well it recognizes its binding sites.

  private void evaluate ()
  {
    for (int b = 0; b < bug.length; ++b) {
      ++bug[b].age;
      bug[b].scoreChromosome(siteInd);
    }
  }

//                      Randomize Scores
//                      --------- ------


  // This method is invoked when selection has been turned off. It
  // randomizes the creatures' sort value (score), effectively eliminating
  // selection.

  private void randomizeSort ()
  {
    for (int b = 0; b < bug.length; ++b) bug[b].sortRandom();
  }

//                      Order Phase
//                      ----- -----


  // This method orders the creatures in the bug array from fewest
  // mistakes to most mistakes

  private void order ()
  {
    Arrays.sort(bug);
  }

}

//                      Creatures
//                      ---------


class Creature implements Comparable
{
  private static final int UNSCORED = Integer.MIN_VALUE;

  public static int idCounter = 0;
  
  private static Param param;         // Parameter object for easy access.
  private static int genePos;         // Gene position.
  private static int thresholdPos;    // Threshold position.
  private static int thresholdWidth;  // Threshold width.
  private static int availablePos;    // First position past gene.
  private static Random generator;    // Random number generator.
  private static int firstNegWeight,  // First weight value that represents
                                      // a negative number.
                     negateWeight;    // Amount to subtract to negate it.
  
  // Each creature has the following information:
  
  public int id;                  // Unique ID.
  public int age = 0;             // Age in generations.
  public byte chromosome[];       // The sequence of bases in its genome.
  public int composition[];       // The count of each of the four bases
                                  // in its chromosome.
  private boolean valid = false;  // Following are valid only if true:
  private int matrix[][];         //   Cached recognizer weight matrix.
  public int threshold;           //   Cached recognizer threshold.
  public int mistakes = UNSCORED; // Latest valuation score, or UNSCORED.
                                  // Following are valid only if scored:
  private int valuation[];        //   Valuation of each position.
  private int spuriousHits = 0;   //   Number of spurious bindings obtained.
  private int sortValue;          //   Value used to sort creatures.
  
  // This constructor is used to create creatures at the beginning
  // of a run. Each creature's chromosome is random.

  public Creature (Param param, int genomeSize, int genePos,
                   int thresholdPos, int thresholdWidth, int availablePos,
                   Random generator)
  {
    // It's a waste to store these static values every time, but it's
    // easier than the alternatives.
    
    this.param = param;
    this.genePos = genePos;
    this.thresholdPos = thresholdPos;
    this.thresholdWidth = thresholdWidth;
    this.availablePos = availablePos;
    this.generator = generator;
    this.firstNegWeight = 1 << (2 * param.getInt(Param.weightWidth) - 1);
    this.negateWeight = firstNegWeight << 1;

    id = nextID();
    chromosome = new byte[genomeSize];
    composition = new int[4];
    matrix = new int[4][param.getInt(Param.siteWidth)];
    valuation = new int[chromosome.length];
    Arrays.fill(valuation, UNSCORED); 

    // Generate a random chromosome and initialize its composition.

    Arrays.fill(composition, 0);
    for (int p = 0; p < chromosome.length; ++p) {
      byte b = Base.random(generator);
      chromosome[p] = b;
      ++composition[b];
    }
  }

  // This constructor is used to create a creature that is a clone of
  // an existing parent.
  
  public Creature (Creature parent)
  {
    id = nextID();
    chromosome  = (byte[])(parent.chromosome.clone());
    composition = (int[])(parent.composition.clone());
    matrix = new int[4][param.getInt(Param.siteWidth)];
    valuation = new int[chromosome.length];
    Arrays.fill(valuation, UNSCORED); 
  }

  // This method compares the scores of two creatures. By implementing
  // the Comparable interface, we can sort creatures using Array.sort().

  public int compareTo (Object c)
  {
    int other = ((Creature)c).sortValue;

    return (this.sortValue < other) ? -1
                                    : ((this.sortValue > other) ? 1 : 0);
  }

  // This method performs a point mutation on the chromosome.

  public void mutate ()
  {
    int p = generator.nextInt(chromosome.length);
    byte old = chromosome[p];
    byte fresh = Base.random(generator);

    if (fresh == old) return;     // No reason to make an identity mutation.

    --composition[old];
    chromosome[p] = fresh;
    ++composition[fresh];
    if (p < availablePos) valid = false;
    mistakes = UNSCORED;
  }

  // This method kills off a creature and replaces it with a clone of a
  // parent creature.
  
  public void killAndClone (Creature parent)
  {
    id = nextID();
    age = 0;
    chromosome = (byte[])(parent.chromosome.clone());
    composition = (int[])(parent.composition.clone());
    valid = false;
    mistakes = UNSCORED;
  }

//                      Creature Weight Matrix
//                      -------- ------ ------


  // This method builds the weight matrix and threshold, if necessary.
  // The matrix is represented in the chromosome as groups of four weights,
  // one per base, repeated times the number of bases per binding site.

  private void extractMatrix ()
  {
    if (valid) return;  // The matrix and threshold are already computed.

    int width = param.getInt(Param.weightWidth);

    for (int p = genePos, s = 0; s < param.getInt(Param.siteWidth); ++s)
      for (int b = 0; b <= 3; ++b) {
        matrix[b][s] = extractWeight(p, width);
        p += width;
      }

    threshold = extractWeight(thresholdPos, thresholdWidth);
    valid = true;
  }

  // This method extracts a weight integer from the chromosome. The
  // weight is computed directly for weight widths in the range 1--5.
  // The other cases are done in a loop. Since a weight is stored in 2's
  // complement, it might have to be negated. 

  private int extractWeight (int pos, int width)
  {
    int weight = 0;

    switch (width) {
      case 1:   weight = chromosome[pos]; break;
      case 2:   weight = (chromosome[pos] << 2) + chromosome[pos+1]; break;
      case 3:   weight = (chromosome[pos] << 4) + (chromosome[pos+1] << 2) +
                         chromosome[pos+2]; break;
      case 4:   weight = (chromosome[pos] << 6) + (chromosome[pos+1] << 4) +
                         (chromosome[pos+2] << 2) + chromosome[pos+3]; break;
      case 5:   weight = (chromosome[pos] << 8) + (chromosome[pos+1] << 6) +
                         (chromosome[pos+2] << 4) + (chromosome[pos+3] << 2) +
                         chromosome[pos+4]; break;
      default:  for (int i = pos; i < pos + width; ++i)
                  weight = (weight << 2) + chromosome[i];
                break;
    }

    if (weight >= firstNegWeight) weight -= negateWeight;
    return weight;
  }

//                      Creature Scoring
//                      -------- -------


  // This method scores a creature's chromosome. As a side effect, it
  // saves the weighting value at each position and counts the number of
  // spurious bindings, which we need to create the PosInfo array.

  public void scoreChromosome (boolean siteInd[])
  {
    if (mistakes != UNSCORED) return;   // Score already calculated.
    extractMatrix();                    // Build weight matrix (maybe).

    int width   = param.getInt(Param.siteWidth),
        missed  = param.getInt(Param.missedSite),
        gene    = param.getInt(Param.spuriousGene),
        nongene = param.getInt(Param.spuriousNongene);
    
    mistakes = spuriousHits = 0;
    for (int p = 0; p <= chromosome.length - width; ++p) {
      if ((valuation[p] = siteValuation(p, width)) >= threshold) {
        if (!siteInd[p]) {
          if (p < availablePos) mistakes += gene; else mistakes += nongene;
          ++spuriousHits;
        }
      } else {
        if (siteInd[p]) mistakes += missed;
      }
    }

    sortValue = mistakes;               // We will sort on the mistake
                                        // count, unless randomized.
  }

  // This method calculates the valuation for the specified binding
  // site. The cases for sites of width 1--7 are performed directly,
  // while wider sites are calculated in a loop. The weight matrix is
  // assumed to be up to date.

  private int siteValuation (int pos, int width)
  {
    byte chr[] = chromosome;
    int weight = 0;

    switch (width) {
    case 1:   weight = matrix[chr[pos  ]][0]; break;
    case 2:   weight = matrix[chr[pos  ]][0] + matrix[chr[pos+1]][1]; break;
    case 3:   weight = matrix[chr[pos  ]][0] + matrix[chr[pos+1]][1] +
                       matrix[chr[pos+2]][2]; break;
    case 4:   weight = matrix[chr[pos  ]][0] + matrix[chr[pos+1]][1] +
                       matrix[chr[pos+2]][2] + matrix[chr[pos+3]][3]; break;
    case 5:   weight = matrix[chr[pos  ]][0] + matrix[chr[pos+1]][1] +
                       matrix[chr[pos+2]][2] + matrix[chr[pos+3]][3] +
                       matrix[chr[pos+4]][4]; break;
    case 6:   weight = matrix[chr[pos  ]][0] + matrix[chr[pos+1]][1] +
                       matrix[chr[pos+2]][2] + matrix[chr[pos+3]][3] +
                       matrix[chr[pos+4]][4] + matrix[chr[pos+5]][5]; break;
    case 7:   weight = matrix[chr[pos  ]][0] + matrix[chr[pos+1]][1] +
                       matrix[chr[pos+2]][2] + matrix[chr[pos+3]][3] +
                       matrix[chr[pos+4]][4] + matrix[chr[pos+5]][5] +
                       matrix[chr[pos+6]][6]; break;
    default:  for (int i = 0; i < width; ++i)
                weight += matrix[chr[pos+i]][i];
              break;
    }
    return weight;
  }

  // This method assigns a random number to the sort value, so that all
  // the creatures will be sorted arbitrariliy.

  public void sortRandom ()
  {
    sortValue = generator.nextInt(1000000);
  }

//                      Creature Min/Max Valuations
//                      -------- ------- ----------


  public int getMinValuation ()
  {
    int siteWidth = param.getInt(Param.siteWidth);

    extractMatrix();                    // Build weight matrix (maybe).

    int min = 0;
    for (int s = 0; s < siteWidth; ++s)
      min += Math.min(matrix[0][s],
                      Math.min(matrix[1][s],
                               Math.min(matrix[2][s], matrix[3][s])));
    return min;
  }

  public int getMaxValuation ()
  {
    int siteWidth = param.getInt(Param.siteWidth);

    extractMatrix();                    // Build weight matrix (maybe).

    int max = 0;
    for (int s = 0; s < siteWidth; ++s)
      max += Math.max(matrix[0][s],
                      Math.max(matrix[1][s],
                               Math.max(matrix[2][s], matrix[3][s])));
    return max;
  }

//                      Creature Position Info
//                      -------- -------- ----


  public PosInfo[] getPosInfo (int generation, boolean siteInd[])
  {
    int weightWidth = param.getInt(Param.weightWidth);
    int siteWidth = param.getInt(Param.siteWidth);
    PosInfo info[] =
      new PosInfo[4 * siteWidth + 1 +
                  ((generation > 0) ? param.getInt(Param.sites) : 0) +
                  spuriousHits];

    extractMatrix();                    // Build weight matrix (maybe).
    int i = -1;

    for (int p = genePos, s = 0; s < siteWidth; ++s)
      for (int b = 0; b <= 3; ++b) {
        info[++i] = new PosInfo(p, weightWidth,
                                PosInfo.GENE_WEIGHT, matrix[b][s]);
        p += weightWidth;
      }
    
    info[++i] = new PosInfo(thresholdPos, thresholdWidth,
                            PosInfo.GENE_THRESHOLD, threshold);

    if (generation > 0) {
      for (int p = 0; p < chromosome.length; ++p) {
        if (siteInd[p])
          info[++i] = new PosInfo(p, siteWidth,
                                  (valuation[p] >= threshold) ?
                                    PosInfo.SITE_MATCHED :
                                    PosInfo.SITE_UNMATCHED,
                                  valuation[p]);
        else if (valuation[p] >= threshold)
          info[++i] = new PosInfo(p, siteWidth,
                                  PosInfo.NONSITE_MATCHED, valuation[p]);
      }
    }

    Arrays.sort(info);
    return info;
  }

//                      Utilities
//                      ---------


  // Return the next creature ID.
  
  private int nextID ()
  {
    return ++idCounter;
  }
}
